(* version 3.2 (c) Copyright 2006-2018 Pierre Corbineau (pour l'essentiel de la premire partie, jusqu' la l. 152 environ), Jean-Baptiste Guillaumin (le reste, par ttonnements successifs...) *)
(* published under the GNU General Public License *)

(*Par rapport  la version prcdente (2): on fait en sorte que les abrviations comptent un peu; ainsi toute abrviation est dveloppe par le script shell, mais on lui ajoute le caractre '#' qui introduit une distance minime (1/8) par rapport aux leons ne comportant pas l'abrviation; mme chose pour les majuscules (identifies par le caractre ^ avant la lettre). 

Le e cdill doit par ailleurs etre reprsent par le caractre @ (et non  comme dans la version prcdente). Dans cette version, on introduit galement le caractre "0", qui permet de forcer ponctuellement une distance nulle par rapport  n'importe quel autre caractre. Voir la page de prsentation et le fichier "lisez-moi" pour plus d'informations. 

Cette version (3.2) fournit galement la matrice des distances au format nexus: le rsultat peut ainsi tre utilis par des logiciels de phylognie.*)

open Sys;;
open Unix;;
open Printf;;

module type Entity =
sig

type t

val poids : t -> int

val dist : t -> t -> int

end

let min a b c = 
 ( let min_ab = if a < b then a else b in
    if c < min_ab then c else min_ab :int)

module Word(E:Entity):Entity with type t = E.t array =
struct

  type t = E.t array

  let dist w1 w2 =
    let hauteur = Array.length w1 in
    let largeur = Array.length w2 in
    let poids1 = Array.map E.poids w1 in
    let poids2 = Array.map E.poids w2 in
    let rangee = Array.make (1+largeur) 0 in
    let diagonale = ref 0 in  
      for i = 0 to largeur-1 do
	let result = rangee.(i) + poids2.(i) in
	  rangee.(i+1)<- result;
      done;
      for j = 0 to hauteur-1 do
	diagonale:= rangee.(0);
	let result = rangee.(0) + poids1.(j) in
	  rangee.(0)<- result;
	  for i = 0 to largeur-1 do
	    let result = min 
            (rangee.(i) + poids2.(i)) 
	    (rangee.(i+1) + poids1.(j))
	    (!diagonale + E.dist w1.(j) w2.(i)) in
	      diagonale:=rangee.(i+1);
	      rangee.(i+1)<- result;
	  done;
      done;
      rangee.(largeur)
	
  let poids w = dist w [||] 

end

module Lettre =
struct
  type t = char  
(*Essai de paramtrage des substitutions*)
  let poids c = 
    if c = 'h' then 4 else 
	(*La prsence ou absence d'un h n'est jamais dterminante. La prsence d'une abrviation est compte trs lgrement.*)
      if c = '#' then 1 else
	(*Le symbole # marque la prsence d'une abrvation. Il est compt trs lgrement, de manire  ce que la prsence d'abrviations systmatiquement au mme endroit fasse sens.*)
	if c = '^' then 1 else
	  (*Le symbole ^ indique que la lettre suivante est crite en capitale. Le phnomne est compt trs lgrement (et diffremment de l'abrviation).*)
	  if c = '0' then 0
	  (*Le symbole 0 peut tre utilis pour forcer une distance nulle par rapport  n'importe quel autre caractre. Cela peut tre utile pour viter d'accrotre trop fortement la distance totale en cas d'omission longue (pour une chane qui varie peu par ailleurs).*)
else 8

let dist c1 c2 =
  if c1 = c2 then 0 else
    match c1,c2 with
(*Variantes graphiques  ne pas prendre en compte*)
    | 'v','u' | 'u','v' -> 0
    | 'i','j' | 'j','i' -> 0
(*Variantes graphiques lgrement prises en compte*)
    | 'e','@' | '@','e' -> 2
    | 'k','c' | 'c','k' -> 2
    | 'i','y' | 'y','i' -> 4
    | 'u','y' | 'y','u' -> 4
    | 'm','n' | 'n','m' -> 4
(*Tilde et nasales*)
    | '-','m' | 'm','-' -> 2
    | '-','n' | 'n','-' -> 2
(*Lettres parfois confondues*)
    | 't','c' | 'c','t' -> 4
    | 'p','b' | 'b','p' -> 6
    | 't','d' | 'd','t' -> 6
    | 'c','g' | 'g','c' -> 6
(*Autres valeurs spcifiques*)
    | '0',_ | _,'0' -> 0   
    | '-',_ | _,'-' -> 4
    | '/',_ | _,'/' -> 4
    | '*',_ | _,'*' -> 4
    | _ -> 8
           
end

module Mot = Word(Lettre)

module Paragraphe =  Word(Mot)

(* module Texte = Word(Paragraphe) *)

(*Dans cette fonction, on peut dfinir l'espace comme caractre si on veut travailler sur une entit de base ("mot") plus large que le mot rel.*)
let is_letter = function
'a'..'z' | 'A'..'Z' | '_' | '@'| (*' ' |*) '-' | '#' | '^' | '*' | '/' | '0' -> true
  | _ -> false

let read_words file =
  let chan = open_in file in
  let word_stack = Stack.create () in
  let buf = Buffer.create 256 in
  try 
    while true do
      let c = input_char chan in
      if is_letter c then
	  Buffer.add_char buf (Char.lowercase c)
            else
	  begin
	    if Buffer.length buf > 0 then
	      Stack.push (Buffer.contents buf) word_stack;
	    Buffer.clear buf
	  end
    done;
    assert false
  with End_of_file ->
    let _ = close_in chan in
    let n_words = Stack.length word_stack in
    let text = Array.make n_words [||] in
      for i=(n_words-1) downto 0 do
	let w = Stack.pop word_stack in
	let wt=Array.init (String.length w) (fun n -> w.[n]) in
	  text.(i)<- wt
      done;
      assert (Stack.length word_stack = 0);
      text


let calcul file1 file2 = 
  let text1 = read_words file1 in
  let text2 = read_words file2 in 
  let dist = Paragraphe.dist text1 text2 in
  let resultat = (float_of_int(dist)/.8.0) in
    resultat
   
let main () =
  let nombre_fichiers = ref ((Array.length Sys.argv) - 1) in
  (*On construit une matrice remplie de 0. correspondant au nombre de fichiers +1 (pour commencer  1)*)
  let matricedist = Array.make_matrix (!nombre_fichiers + 1) (!nombre_fichiers + 1) 0. in

  (*On remplit la matrice  l'aide de la fonction "calcul" pour chaque paire de fichiers.*)
  for j = 1 to !nombre_fichiers do
    for i = j+1 to !nombre_fichiers do
      let res = calcul Sys.argv.(i) Sys.argv.(j) in
        matricedist.(i).(j) <- res;
      matricedist.(j).(i) <- res;
    done; 
  done;

(*Les lignes suivantes, jusqu' la mention "Fin de l'ajout", permettent d'diter la matrice des distances au format nexus. La cration d'un fichier .nex est ensuite prise en charge par le script calcul.*)
  Printf.printf "#nexus\n\n begin taxa;\n  dimensions ntax=%i;\n  taxlabels " !nombre_fichiers;

 for j = 1 to !nombre_fichiers do
   Printf.printf "%s " Sys.argv.(j);
 done;

  Printf.printf ";\nend;\n\nbegin distances;\n  dimensions ntax=%i;\nformat diagonal labels triangle=lower;\n  matrix"  !nombre_fichiers;

 for j = 1 to !nombre_fichiers do
   Printf.printf "\n%s " Sys.argv.(j);
   for i = 1 to j do
     Printf.printf "%f " matricedist.(j).(i);
   done;
 done;

 Printf.printf ";\nend;\n\n";
(*Fin de l'ajout.*)

(*on attribue le n d'argv  chaque ligne/colonne (matricedist.(0).(i) et (i).(0)) pour pouvoir s'y retrouver ensuite*)

for i = 0 to !nombre_fichiers do
  matricedist.(0).(i) <- float_of_int(i);
  matricedist.(i).(0) <- float_of_int(i);
done;

(*On copie Sys.argv dans elements, pour pouvoir travailler ensuite sur ce tableau*)

let elements = ref Sys.argv in

(*On garde en mmoire la moyenne de chaque paire d'lments traite, afin de la sousrtaire au moment o on calcule la distance entre deux noeuds. On l'initialise  zro (ainsi les moyennes d'lments non encore traits ne sont pas prises en compte).*)

let moyenne = Array.make (!nombre_fichiers + 1) 0. in

(*on calcule la somme des longueurs des branches, totalbranches, d'aprs la formule de l'algorithme NJ, en choisissant successivement toutes les paires i, j possibles. On stocke les rsultats dans une matrice.*)

let totalbranches = Array.make_matrix (!nombre_fichiers +1) (!nombre_fichiers + 1) 0. in

let choix = ref [|0;0|] in

for etapes = !nombre_fichiers downto 3 do (*On s'arrte  3:  l'issue du processus, il reste un seul tmoin, avec sa distance au noeud form par tous les autres.*)

(*La valeur de minimum va tre raffecte  la baisse au cours de la boucle; on la fixe pour l'instant trs haut (on pourrait raffiner davantage...).*)

let minimum = ref (100000.) in

for j = 1 to !nombre_fichiers do
  for i = 1 to j-1 do

    let somme1 = ref (0.) in
    let somme2 = ref (0.) in
    
      (*somme1 = somme des distances i-k+j-k pour i,j  diffrents de k
      somme2 = somme des distances a-b pour a et b diffrents de i et j*)
    
(*      for k = 1 to !nombre_fichiers do
        if k != i then
          somme1 := !somme1 +. matricedist.(i).(k);
        if k!= j then
        somme2 := !somme2 +. matricedist.(j).(k); *)

    for k = 1 to !nombre_fichiers do
      if (k != i && k!= j) then
      somme1 := !somme1 +. matricedist.(i).(k) +. matricedist.(j).(k);
  	if (k!= i && k!= j) then
	  for b = 1 to k-1 do
	    if (b!= i && b!= j) then
	      somme2 := !somme2 +. matricedist.(k).(b);
	  done;
        done;

	  (*	      let somme = (float_of_int(!nombre_fichiers)-.2.)*.matricedist.(i).(j) -. !somme1 -. !somme2 in *)

	  (* On calcule la somme correspondant  la longueur totale des embranchements pour i et j, et on la place dans la matrice totalbranches.*)
    
	  let somme = (1./.(2.*.(float_of_int(!nombre_fichiers)-.2.)))*. !somme1+.(1./.2.)*.matricedist.(i).(j)+.(1./.(float_of_int(!nombre_fichiers)-.2.))*. !somme2 in

          totalbranches.(i).(j) <- somme;

      (*Si cette somme est infrieure  la somme minimale dj constate, elle la remplace. La valeur de i et de j est stocke dans le tableau "choix".*)
	  
      if somme < !minimum then
        begin
        minimum := somme;
        choix := [|i;j|];
        end
  done;
done;

(*Une fois qu'on a test tous les i et j, et dtermin la paire qui minimise la longueur totale des embranchements, on calcule la distance des tmoins retenus par rapport  leur noeud commun.*)

    let somme3 = ref (0.) in
    let somme4 = ref (0.) in

for k = 1 to !nombre_fichiers do
  if k != !choix.(0) && k != !choix.(1) then 
    begin
      somme3 := !somme3 +. matricedist.(!choix.(0)).(k);
      somme4 := !somme4 +. matricedist.(!choix.(1)).(k);
    end 
done;

(*distance1 est la distance du noeud par rapport au premier tmoin retenu; distance2 par rapport au second.*)
    
let distance1 = ((matricedist.(!choix.(0)).(!choix.(1))) +. (!somme3/.(float_of_int(!nombre_fichiers) -.2.)) -. (!somme4/.(float_of_int(!nombre_fichiers) -.2.)))/.2. in

(* let distance2 = ((matricedist.(!choix.(0)).(!choix.(1))) +. (!somme4/.(float_of_int(!nombre_fichiers) -.2.)) -. (!somme3/.(float_of_int(!nombre_fichiers) -.2.)))/.2. in *)

let distance2 = matricedist.(!choix.(0)).(!choix.(1)) -. distance1 in

(*Quand on doit calculer la distance du nouveau noeud  un noeud tabli pralablement, on considre la distance  ce dernier comme la moyenne des distances aux lments qui le forment. D'o la matrice "moyenne", qui garde une trace de ces donnes des itrations prcdentes.*)

let anciennemoyenne1 = moyenne.(!choix.(0)) in
let anciennemoyenne2 = moyenne.(!choix.(1)) in

let distancenoeuds1 = distance1-.anciennemoyenne1 in
let distancenoeuds2 = distance2-.anciennemoyenne2 in

(*On redfinit le tableau "moyenne" en supprimant la colonne choix.(1) et en dcalant les suivantes (de mme que dans l'tape "reduction step" prsente ci-dessous pour la matrice des distances).*)

moyenne.(!choix.(0)) <- (matricedist.(!choix.(0)).(!choix.(1)))/.2.;
for c = !choix.(1) to !nombre_fichiers-1 do
  moyenne.(c) <- moyenne.(c+1);
done;

(*On enregistre le noeud et les distances calcules dans la colonne du tableau elements qui correspond au premier choix (!choix.(0)).*)

if !nombre_fichiers == 3 then 
!elements.(int_of_float(matricedist.(!choix.(0)).(0))) <- "(" ^ !elements.(int_of_float(matricedist.(!choix.(0)).(0))) ^ ":" ^ string_of_float(distancenoeuds1) ^ "," ^ !elements.(int_of_float(matricedist.(!choix.(1)).(0))) ^ ":" ^ string_of_float(distancenoeuds2)

else
!elements.(int_of_float(matricedist.(!choix.(0)).(0))) <- "(" ^ !elements.(int_of_float(matricedist.(!choix.(0)).(0))) ^ ":" ^ string_of_float(distancenoeuds1) ^ "," ^ !elements.(int_of_float(matricedist.(!choix.(1)).(0))) ^ ":" ^ string_of_float(distancenoeuds2) ^ ")";


(*"Reduction step": on place la paire choisie  la place du premier choix (!choix.(O)). On supprime le deuxime choix (!choix.(1)) et on dcale les colonnes suivantes de la matrice.*)

nombre_fichiers := !nombre_fichiers -1;

let matricereduction = Array.make_matrix (!nombre_fichiers + 1) (!nombre_fichiers + 1) 0. in

for i = 0 to !nombre_fichiers do
  for j = 0 to i-1 do
    if i < !choix.(0) then  (*Si i (donc j aussi) tombe "avant" le premier lment choisi, on ne change rien*)
      begin
        matricereduction.(i).(j) <- matricedist.(i).(j);
      end 
    else if i == !choix.(0) then  (*Si i correspond au premier lment choisi: si j vaut 0 (qui renvoie  l'identification numrique de l'lment trait, et non  une valeur), on ne change rien. Sinon, on affecte  la matrice la valeur de la moyenne des distances de l'lment j aux deux lments choisis. Dans la colonne du premier lment choisi, la boucle remplace ainsi les distances initiales par les nouvelles distances, aprs regroupement.*) 
      begin
        if j == 0 then matricereduction.(i).(j) <- matricedist.(i).(j) else
        matricereduction.(i).(j) <- (matricedist.(j).(!choix.(0)) +. matricedist.(j).(!choix.(1)))/.2.;
      end 
    else if i > !choix.(0) && i < !choix.(1) then  (*Si i tombe "entre" les deux lments choisis, soit j correspond au premier lment choisi, et on affecte  la colonne en question la moyenne des distances; soit j ne correspond pas au premier lment choisi, et on ne change rien.*)
      begin
        if j == !choix.(0) then matricereduction.(i).(j) <- (matricedist.(i).(!choix.(0)) +. matricedist.(i).(!choix.(1)))/.2. else
        matricereduction.(i).(j) <- matricedist.(i).(j);
      end
    else if i >= !choix.(1) then  (*Si i tombe "aprs" le deuxime lment choisi, plusieurs cas: soit j correspond au premier lment choisi, et on affecte  la colonne la moyenne des distances; soit j est "avant" le deuxime lment choisi, et on "dcale" seulement i (la colonne du deuxime lment choisi doit disparatre); soit j tombe aussi "aprs" le deuxime lment choisi et on "dcale" tout.*)
      begin
        if j == !choix.(0) then matricereduction.(i).(j) <- (matricedist.(i+1).(!choix.(0)) +. matricedist.(i+1).(!choix.(1)))/.2.
        else if j < !choix.(1) then
        matricereduction.(i).(j) <- matricedist.(i+1).(j)
        else if j >= !choix.(1) then
            matricereduction.(i).(j) <- matricedist.(i+1).(j+1)
      end;
      matricereduction.(j).(i) <- matricereduction.(i).(j);
 done;
 done;


(*On copie matricereduction dans matricedist et on relance l'algorithme*)

for i = 0 to !nombre_fichiers do
  for j = 0 to i-1 do
    matricedist.(i).(j) <- matricereduction.(i).(j);
    matricedist.(j).(i) <- matricedist.(i).(j);
  done;
done;

done;  (*Fin de la boucle lance par "for etapes downto 3": on reprend tout le processus avec un tmoin en moins*)

(*On ajoute ici la valeur qui reste  la fin de la boucle*)

let distancenoeuds = matricedist.(1).(2)-.moyenne.(1)-.moyenne.(2) in


!elements.(!choix.(0)) <- !elements.(int_of_float(matricedist.(!choix.(0)).(0))) ^ "," ^ (!elements.(int_of_float(matricedist.(0).(2)))) ^ ":" ^ (string_of_float(distancenoeuds)) ^ ");\n";

printf "%s" (!elements.(!choix.(0)));

    exit 0;;
main ();;

